/* Routine to compute tomographic projection matrix

   This routine does not use line integrals, but rather
   a projection shape as defined in the subroutine
   'proj_shape' in this file. This is set to 
   a TRIANGULAR profile right now with width PRJ_WIDTH.
   
   Based on this shape, approximate line integrals
   through the pixels are pre-computed for different
   angles ('profiles').    
   This computation is performed by subsampling
   each pixel by the factor 'PIX_SUBSAMPLE' in each
   direction and calculating the sum of proj-shape value at the
   center of the subsamples times the area of
   the subsamples.
   
   In this version, A-matrix is computed using doubles 
   only, but then stored using floats. When returning
   a single column of the matrix, the stored values
   are first converted to doubles and then returned. 
   If the program is used without the store option, 
   then everything is computed as doubles. 

   Furthermore, the row indices are stored using 
   UNSIGNED SHORT to save storage space. This limits 
   the number of projections to 65536, if more 
   projections are needed, change all 'unsigned short'
   to 'unsigned int' in this file and remove the
   error message. No changes to any other files
   will be necessary.
   
*/


/* 8/21/00: Added freeing of A-matrix for specific
   resolution levels by calling A_comp(-1,-1,...,res_level); */


#include <stdio.h>
#include <math.h>
#include <mat.h>
#include <matrix.h>
#include <stdlib.h>
#include "A_comp_dbl_storeflt.h"
#include "allocate.h"

#define LEN 512
#define EVEN 0
#define ODD 1
#define MAX_RES_LEVELS 20

#define PROJ_MLT 1.0      /* amplitude scaling */
#define PRJ_WIDTH 2.34    /* in mm */

#define PIX_SUBSAMPLE 15 /* for projection profiles */



void A_comp_ij(
	    int im_row, int im_col,
	    struct sino_info *sino,
	    struct geom_info *geom,
	    double **pix_prof,
	    struct column *sp_col);

static void A_coarse(int im_row, int im_col,struct sino_info *sino,struct geom_info *geom,double **pix_prof,struct column *sp_col,int res_level);

double proj_shape(double delta_t);



double proj_shape(double delta_t)
{
  double val;
      
  if (delta_t<0.0)
    delta_t = -delta_t;

  if (delta_t>=PRJ_WIDTH/2.0)
    val=0;
  else
    {
      val = (1.0 - 2.0*delta_t/PRJ_WIDTH)/PRJ_WIDTH*(double)PROJ_MLT;

    }

  return(val);
}



void pix_prof_comp(struct sino_info *sino,struct geom_info *geom,double ***pix_prof)
{

  int	i,j,ang_index,t_index;
  double Fdbl,delta_pix,incr_delta_t,ang,pi;
  double delta_m,csa,ssa,xy0_m,A_m,x_m,y_m;
  double t_prj,sum,t_m,addp;
  double max_delta_t;

  /*FILE *fp;*/

  *pix_prof = (double **)get_img(LEN,sino->n_theta,sizeof(double));
 
  pi = 4.0 * atan(1.0);

  Fdbl=(double)PIX_SUBSAMPLE;
	
  delta_pix = geom->delta;

  max_delta_t = delta_pix/sqrt(2.0) + 0.5*PRJ_WIDTH;
  sino->max_delta_t = max_delta_t;

  incr_delta_t = max_delta_t/(double)LEN;


  printf("  delta_pix = %f mm\n  projection_width = %f mm\n",delta_pix,PRJ_WIDTH);
  /* m denotes the micropixels used to subsample a pixel */
  delta_m = delta_pix/Fdbl;
  xy0_m = -(Fdbl-1.0)/2.0 * delta_m;
  A_m = (delta_pix/Fdbl)*(delta_pix/Fdbl); /* area of micropix */

  for (ang_index=0;ang_index<sino->n_theta;ang_index++)
    {
  
      ang = sino->theta_0 + sino->delta_theta * (double)ang_index;

      while(ang >= pi/2.0)
	ang -= pi/2.0;
      
      csa=cos(ang);
      ssa=sin(ang);

      for (t_index=0;t_index<LEN;t_index++)
	{
	  t_prj = incr_delta_t * (double)t_index;
	  sum=0.0;

	  for (i=0;i<PIX_SUBSAMPLE;i++)
	    {
	      y_m = xy0_m + delta_m*(double)i;
	      for (j=0;j<PIX_SUBSAMPLE;j++)
		{
		  x_m = xy0_m + delta_m*(double)j;
		  t_m = y_m*csa-x_m*ssa;
		  /*printf("x_m=%f y_m=%f t_m=%f\n",x_m,y_m,t_m);*/
		  addp = proj_shape(t_prj-t_m);
		  sum += addp;
		}
	    }

	  sum *= A_m;

	  (*pix_prof)[ang_index][t_index] =sum;
	}
    }
  /*
    fp=fopen("prof","w");
    for (ang_index=0;ang_index<sino->n_theta;ang_index++)
      {
        for (t_index=0;t_index<LEN;t_index++)
        fprintf(fp,"%g ",(*pix_prof)[ang_index][t_index]);
        fprintf(fp,"\n");
      }
    fclose(fp);
   */ 
} /* Close prof_comp()	*/




/* Comments added 3/3/98 TF */
/* Acomp computes one column of the A-matrix, i.e. 
   takes pixel location row,col as input and returns 
   sp_col , where int *spcol.row is a list of the 
   projection indices (proj. in 1D order (angle*n_t + t)) 
   that intersect with the pixel. 
   double  sp_col.val[i]  contains the overlap of the 
   intersection with projection spcol.row[i] 
   spcol.n_row is number of projections that intersect 
   with pixel 
*/
void A_comp(
	    int im_row, int im_col,
	    struct sino_info *sino,
	    struct geom_info *geom,
	    double **pix_prof,
	    struct column *sp_col, int res_level)
{

#ifdef STORE_A_MATRIX
  /* precompute entire A-matrix at first call, then 
     just return precomputed cols in subsequent calls */

  static int first=1;
  static struct column sp_col_sgl;
  static struct column_flt ***sp_col_arr,*spc_f;
  static int level_FLAG[MAX_RES_LEVELS];
  int i,j,r,nr;


  if (sino->n_t*sino->n_theta > 65536)
    {
      fprintf(stderr,"Error: A_comp uses unsigned short for row array, can't have more than 65536 projections, in your case sino->n_t*sino->n_theta=%d.\nChange all unsigned short to unsigned int in the file ansi_lib_loc/A_comp_storeflt.c and it will work.\n\a\n",sino->n_t*sino->n_theta);
      exit(-1);
    }


  if (res_level>=MAX_RES_LEVELS)
    {
      fflush(stdout);
      fprintf(stderr,"Error in A_comp: resolution level=%d > %d=MAX_RES_LEVELS\n",res_level,MAX_RES_LEVELS);
      exit(-1);
    }

  if (first==1)
    {
      first=0;fflush(stderr);
      

      sp_col_sgl.val = (double *) get_spc(sino->n_t*sino->n_theta, sizeof(double));
      sp_col_sgl.row = (int *) get_spc(sino->n_t*sino->n_theta, sizeof(int));

      sp_col_arr=(struct column_flt ***)get_spc(MAX_RES_LEVELS, sizeof(struct column_flt **));
      
      /* mark all resolution levels as not already computed */
      for (i=0;i<MAX_RES_LEVELS;i++)
	level_FLAG[i]=1;
    }


  if ((im_row==-1) && (im_col==-1))
    {
      printf("Freeing projection matrix for scale %d\n",res_level);
      if (level_FLAG[res_level]) 
	 {
	   fprintf(stderr,"Can't free matrix for scale %d since never allocated.\n\a\n",res_level);
	   exit(-1);
	 }
      for (i=0;i<geom->n_y;i++)
	for (j=0;j<geom->n_x;j++)
	  {
	    /*printf("res_level=%d i=%d, j=%d\n",res_level,i,j);*/

	    free((void*)sp_col_arr[res_level][i][j].val);
	    free((void*)sp_col_arr[res_level][i][j].row);
	  }
      multifree((void*)sp_col_arr[res_level],2);
      level_FLAG[res_level]=1;
    }
  else
    {
      if (level_FLAG[res_level])
	{
	  printf("Precomputing projection matrix for scale %d\n",res_level);
	  fflush(stdout);fflush(stderr);
	  level_FLAG[res_level]=0;
	  
	  if (res_level>0)
	    for (i=0;i<res_level;i++)
	      if (level_FLAG[i])
		{
		  fprintf(stderr,"Error: First need to compute A-matrix for all smaller scales. scale %d not yet computed.\n\a\n",i);
		  exit(-1);
		}
	  
	  sp_col_sgl.multiplier = 1.0; 
	  sp_col_sgl.A_cor = sp_col->A_cor;
	  
	  
	  sp_col_arr[res_level]=(struct column_flt **)multialloc(sizeof(struct column_flt),2,geom->n_y,geom->n_x);
	  
	  
	  for (i=0;i<geom->n_y;i++)
	    for (j=0;j<geom->n_x;j++)
	      {
		if (res_level==0)
		  A_comp_ij(i,j,sino,geom,pix_prof,&sp_col_sgl);
		else
		  A_coarse(i,j,sino,geom,pix_prof,&sp_col_sgl,res_level);
		
		
		
		sp_col_arr[res_level][i][j].val=(float *) get_spc(sp_col_sgl.n_row, sizeof(float));
		sp_col_arr[res_level][i][j].row = (unsigned short *) get_spc(sp_col_sgl.n_row, sizeof(unsigned short));
		sp_col_arr[res_level][i][j].n_row = sp_col_sgl.n_row;
		sp_col_arr[res_level][i][j].multiplier = sp_col_sgl.multiplier;
		for (r=0;r<sp_col_sgl.n_row;r++)
		  {
		    sp_col_arr[res_level][i][j].val[r] = (float)sp_col_sgl.val[r];
		    sp_col_arr[res_level][i][j].row[r] = (unsigned short)sp_col_sgl.row[r];
		  }

	      }
	  /*free((void*)sp_col_sgl.val);
	    free((void*)sp_col_sgl.row);*/
	  fflush(stderr);
	  fflush(stdout);
	}
      
      spc_f=&sp_col_arr[res_level][im_row][im_col];
      sp_col->n_row = nr = spc_f->n_row;
      sp_col->multiplier = spc_f->multiplier;
      
      for (r=0;r<nr;r++)
	{
	  sp_col->val[r]=(double)spc_f->val[r];
	  sp_col->row[r]=(int)spc_f->row[r];
	}
    }

#else
  /* compute A_*j on the fly */
  if (res_level==0)
    A_comp_ij(im_row,im_col,sino,geom,pix_prof,sp_col);
  else
    A_coarse(im_row, im_col, sino,geom,pix_prof,sp_col,res_level);
#endif
}




void A_comp_ij(
	    int im_row, int im_col,
	    struct sino_info *sino,
	    struct geom_info *geom,
	    double **pix_prof,
	    struct column *sp_col)
{
  int	ind,ind_min,ind_max,pr,prof_ind,pind,i,proj_count;

  double Aval,t_min,t_max,ang,x,y,t_pix_ctr,t_prj,Dt;
  double max_delta_t;
  static int first=1;
  static double *cs,*si;

  if (first)
    {
      /* precompute cos and sin for all projection angles*/
      first=0;
      cs=(double*)mget_spc(sino->n_theta,sizeof(double));
      si=(double*)mget_spc(sino->n_theta,sizeof(double));
      
      for (pr=0; pr < sino->n_theta; pr++)
	{
	  ang = sino->theta_0 + pr*(sino->delta_theta);
	  cs[pr]=cos((double) ang);
	  si[pr]=sin((double) ang);
	}
    }


  max_delta_t=sino->max_delta_t;
  y = geom->y_0 + im_row*geom->delta;
  x = geom->x_0 + im_col*geom->delta;
  
  proj_count = 0;
  for(pr=0; pr < sino->n_theta; pr++)
    {

      /*ang = sino->theta_0 + pr*(sino->delta_theta);*/
      pind = pr*sino->n_t;
      
      /* Need profile to contain 2 pixel widths       */
      /* pixel center is    y*cos((double) ang) - x*sin((double) ang) */
      /* so go down to tmin=center-1  and upto  t_max-center+1 */
      /* consistent with profiles computation t=0..2 */
      /*t_min = y*cos((double) ang) - x*sin((double) ang) - geom->delta;
	t_max = t_min + 2.0*geom->delta;*/
      

      t_pix_ctr = y*cs[pr] - x*si[pr];
      t_min = t_pix_ctr - max_delta_t;
      t_max = t_pix_ctr + max_delta_t;
      
      /* Check whether pixel is hit at all 
	 or t-range below even the first parallel proj. */
      if(t_max < sino->t_0) /* This also prevents over-reach (with rounding of
			       negative numbers)	*/
	continue;
      
      /* From physical t_min and t_max, compute the
	 index range of parallel proj. */
      /* A more general implementation: T.F. May 7, 99 */
      ind_min = ceil((t_min-sino->t_0)/sino->delta_t); 
      ind_max = (t_max-sino->t_0)/sino->delta_t;

  

      /* Fix this 4/91 to prevent over-reach at ends  */
      /* clip ends of t-range */
      ind_min = (ind_min<0) ? 0:ind_min;
      ind_max = (ind_max>=sino->n_t) ? sino->n_t-1:ind_max;
      for(i = ind_min; i <=ind_max; i++) 
	{
	  
	  ind = pind+i; /* linear index in 1D array of all proj. (angle*n_t + t)*/
	  
	  /* get profile_index that is closest to this projection */
	  /* profiles are in discrete t-steps of 2pixels/LEN = 
	     2.0*geom->delta / LEN */
	  t_prj = sino->t_0+i*sino->delta_t;
	  Dt = fabs(t_prj-t_pix_ctr);

	  prof_ind = (int)( Dt/max_delta_t * (double)LEN + 0.5 ); 
	  


	  
	  if(prof_ind >= LEN || prof_ind <0)
	    {
	      if(prof_ind == LEN)
		prof_ind = LEN-1;
	      else if (prof_ind == -1)
		prof_ind = 0;
	      else
		{
		  fprintf(stderr,"Error: Exiting w/ error in Acomp_dbl_storeflt.c.\nprof_ind=%d out of range 0..%d",prof_ind,LEN-1);
		  exit(-1);
		}
	    }
	  /* Assign output */
	  if((Aval=pix_prof[pr][prof_ind]) > 0.0) 
            {
              if (sp_col->A_cor.mode=='t')
	        sp_col->val[proj_count] = Aval;
              else
                if (sp_col->A_cor.mode=='e') 
		  /* in emis-mode, divide by forward proj. trn-sinogram */
		  /* to account for attenuation. */            
		  {
		    sp_col->val[proj_count] = Aval*(sp_col->A_cor.Y_t[pr][i]);
		  }		
		else
		  {
		    fprintf(stderr,"A_comp Error: Unknown mode %c\n\n",sp_col->A_cor.mode);
		    exit(-1);
		  }
	      sp_col->row[proj_count] = ind;/* Index for ray    */
	      proj_count++;
	    }
	}
    } 
  sp_col->n_row = proj_count;

  
  
}/* Close A_comp()	*/




static void A_coarse(int img_row, int img_col,struct sino_info *sino,struct geom_info *geom,double **pix_prof,struct column *sp_col,int res_level)
{
  /*static int first=1;*/
  struct column sp_col_A[4];
  int i,p;
  int row_ind_A[4];
  int row_ind_wav;
  int act_pix,act_prj;
  double act_prj_val;

  /*if (first)
    {
    printf("Allocating multiresolution memory for A-matrix...\n");fflush(stdout);
    first=0;
  */
  for (i=0;i<4;i++)
    {
      sp_col_A[i].val = (double *) get_spc(sino->n_t*sino->n_theta, sizeof(double));
      sp_col_A[i].row = (int *) get_spc(sino->n_t*sino->n_theta, sizeof(int));
    }
  /*  printf("done.\n");fflush(stdout);fflush(stderr);
      }
  */
  for (i=0;i<4;i++)
    {
      sp_col_A[i].multiplier=1.0;
      sp_col_A[i].A_cor=sp_col->A_cor;
    }

  A_comp(2*img_row  ,2*img_col  ,sino,geom,pix_prof,&sp_col_A[0],res_level-1);
  A_comp(2*img_row  ,2*img_col+1,sino,geom,pix_prof,&sp_col_A[1],res_level-1);
  A_comp(2*img_row+1,2*img_col  ,sino,geom,pix_prof,&sp_col_A[2],res_level-1);
  A_comp(2*img_row+1,2*img_col+1,sino,geom,pix_prof,&sp_col_A[3],res_level-1);

  /* Now we compute the columns of the proj. matrix Q for the
     3 wavelet coeffs. */
  for (p=0;p<4;p++)
    row_ind_A[p]=0;

  row_ind_wav=0;

  
  while ((row_ind_A[0]<sp_col_A[0].n_row) ||(row_ind_A[1]<sp_col_A[1].n_row) || (row_ind_A[2]<sp_col_A[2].n_row) ||(row_ind_A[3]<sp_col_A[3].n_row))
    {

      /* act_pix = argmin_p{ sp_col[p].row[ row_ind_A[p] ] }
	 act_prj = min_p{ sp_col[p].row[ row_ind_A[p] ] }  */
      act_pix=-1;
      act_prj=-1;
      for (p=0;p<4;p++)
	if (row_ind_A[p] < sp_col_A[p].n_row)
	  if ((sp_col_A[p].row[row_ind_A[p]] < act_prj) || (act_prj==-1))
	    {
	      act_pix=p;
	      act_prj=sp_col_A[p].row[row_ind_A[p]];
	    }

      /* integrity check */
      if ((act_prj==-1) || (act_pix==-1) || (act_prj >= sino->n_t*sino->n_theta))
	{
	  fprintf(stderr,"error in wav_Acomp: atc_prj/act_pix computation wrong\n\a\n");
	  exit(-1);
	}


      act_prj_val = sp_col_A[act_pix].val[row_ind_A[act_pix]];

      sp_col->row[row_ind_wav] = act_prj;
      sp_col->val[row_ind_wav] = act_prj_val;
	
      
      /* now check if there are pixels p other than act_pix
	 that intersect same projection act_prj */
      /* since min search above is stable would really only need
	 to ckeck for p>act_pix */
      for (p=0;p<4;p++)
	if (p!=act_pix)  
	  if (row_ind_A[p] < sp_col_A[p].n_row)
	    {
	      if (sp_col_A[p].row[row_ind_A[p]] == act_prj)
		{
		  act_prj_val = sp_col_A[p].val[row_ind_A[p]];
		  sp_col->val[row_ind_wav] += act_prj_val;
		  row_ind_A[p]++;
		}
	      else
		if (sp_col_A[p].row[row_ind_A[p]] < act_prj)
		  {
		    fprintf(stderr,"Error in A_ij_coarse, can't be...\n\a\n");
		    exit(-1);
		  }
	    } 
      
      /* IMPORTANT::::::::::: MULTIPLIER FOR WAVELET TRANSFORM ****/
      sp_col->val[row_ind_wav] *= 0.5;

      row_ind_A[act_pix]++;
      row_ind_wav++;
    }

  /* set number of rows for wavelet sp_cols (sparse matrix format) */
  sp_col->n_row = row_ind_wav;

  /* integrity check: all row_ind_A[p] should equal sp_col_A[p].n_row */
  p=0;
  if (row_ind_A[p]!=sp_col_A[p].n_row)
    {
      fprintf(stderr,"Error in A_ij_coarse: row_ind_A[%d]=%d != %d = sp_col_A[%d].n_row\n\a\n",p,row_ind_A[p],sp_col_A[p].n_row,p);
      exit(-1);
    }

  /* Also, if row_ind_wav > sino->n_t*sino->n_theta, then something is 
     seriously wrong */
  if (row_ind_wav > sino->n_t*sino->n_theta)
    {
      fprintf(stderr,"Error in A_ij_coarse: row_ind_wav=%d > sino->n_t*sino->n_theta=%d\n\a\n",row_ind_wav, sino->n_t*sino->n_theta);
      exit(-1);
    }


  for (i=0;i<4;i++)
    {
      free((void*)sp_col_A[i].val);
      free((void*)sp_col_A[i].row);
    }


}









#undef LEN_Z 


void read_mat_sino_info(
  MATFile *fin,       /* pointer to file */
  struct sino_info *sino
  )
{
 mxArray *pm;

 pm = matGetArray(fin,"n_theta");
 if(pm != NULL){
  sino->n_theta = (int) mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"n_t");
 if(pm != NULL){
  sino->n_t = (int) mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"delta_theta");
 if(pm != NULL){
  sino->delta_theta = mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"delta_t");
 if(pm != NULL){
  sino->delta_t = mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"theta_0");
 if(pm != NULL){
  sino->theta_0 = mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 pm = matGetArray(fin,"t_0");
 if(pm != NULL){
  sino->t_0 = mxGetScalar(pm);
  }
 mxDestroyArray(pm);

 sino->dec_a = 1;
 sino->dec_t = 1;

 /*if(VERBOSE)
   {*/

 printf("\nSinogram parameters from input file:\n");
 printf("  delta_t = %f\n",sino->delta_t);
 printf("  delta_theta = %f\n",sino->delta_theta);
 printf("  t_0 = %f\n",sino->t_0);
 printf("  theta_0 = %f\n",sino->theta_0);
 printf("  n_t = %d\n",sino->n_t);
 printf("  n_theta = %d\n\n",sino->n_theta);
 /*}*/

}


